from __future__ import annotations

import os
from functools import lru_cache
from typing import TYPE_CHECKING, Any

from lilya._internal._module_loading import import_string
from lilya.conf.functional import LazyObject, empty

if TYPE_CHECKING:
    from lilya.conf.global_settings import Settings

ENVIRONMENT_VARIABLE = "LILYA_SETTINGS_MODULE"


@lru_cache
def reload_settings() -> type[Settings]:
    """
    Reloads the global settings.
    """
    settings_module: str = os.environ.get(
        ENVIRONMENT_VARIABLE, "lilya.conf.global_settings.Settings"
    )
    settings: type[Settings] = import_string(settings_module)
    return settings


class LilyaLazySettings(LazyObject):
    """
    A lazy proxy for either global Lilya settings or a custom settings object.
    The user can manually configure settings prior to using them. Otherwise,
    Lilya uses the settings module pointed to by LILYA_SETTINGS_MODULE.
    """

    def _setup(self, name: str | None = None) -> None:
        """
        Load the settings module pointed to by the environment variable. This
        is used the first time settings are needed, if the user hasn't
        configured settings manually.
        """
        settings_module: str = os.environ.get(
            ENVIRONMENT_VARIABLE, "lilya.conf.global_settings.Settings"
        )

        settings: type[Settings] = import_string(settings_module)

        for setting, _ in settings().dict().items():
            assert setting.islower(), f"{setting} should be in lowercase."

        self._wrapped = settings()

    def configure(self, override_settings: type[Settings]) -> None:
        """
        Making sure the settings are overriden by the settings_module
        provided by a given application and therefore use it as the application
        global.
        """
        self._wrapped = override_settings

    def __repr__(self: LilyaLazySettings) -> str:
        # Hardcode the class name as otherwise it yields 'Settings'.
        if self._wrapped is empty:
            return "<LilyaLazySettings [Unevaluated]>"
        return f'<LilyaLazySettings "{self._wrapped.__class__.__name__}">'

    @property
    def configured(self) -> Any:
        """Return True if the settings have already been configured."""
        return self._wrapped is not empty


settings: type[Settings] = LilyaLazySettings()  # type: ignore
__lazy_settings__: type[Settings] = LilyaLazySettings()  # type: ignore



from __future__ import annotations

import functools
import inspect
import sys
import typing
from abc import ABC, abstractmethod
from urllib.parse import urlencode

from lilya.compat import is_async_callable
from lilya.exceptions import HTTPException
from lilya.requests import Connection, Request
from lilya.responses import RedirectResponse
from lilya.websockets import WebSocket

if sys.version_info >= (3, 10):  # pragma: no cover
    from typing import ParamSpec
else:  # pragma: no cover
    from typing_extensions import ParamSpec


P = ParamSpec("P")

AuthResult = tuple["AuthCredentials", "UserInterface"]


def has_required_scope(conn: Connection, scopes: typing.Sequence[str]) -> bool:
    """
    Check if the connection has all the required scopes.

    Args:
        conn (Connection): The connection object containing authentication details.
        scopes (typing.Sequence[str]): A sequence of required scope strings.

    Returns:
        bool: True if all required scopes are present in the connection's scopes, False otherwise.
    """
    conn_scopes = set(conn.auth.scopes)
    return all(scope in conn_scopes for scope in scopes)


def requires(
    scopes: str | typing.Sequence[str],
    status_code: int = 403,
    redirect: str | None = None,
) -> typing.Callable[[typing.Callable[P, typing.Any]], typing.Callable[P, typing.Any]]:
    """
    Decorator to enforce required scopes on a function.

    Args:
        scopes (str | typing.Sequence[str]): A single scope or a sequence of required scopes.
        status_code (int, optional): The HTTP status code to return if the scope check fails. Defaults to 403.
        redirect (str | None, optional): The URL to redirect to if the scope check fails. Defaults to None.

    Returns:
        typing.Callable[[typing.Callable[P, typing.Any]], typing.Callable[P, typing.Any]]: The decorated function.
    """
    scopes_list = [scopes] if isinstance(scopes, str) else list(scopes)

    def decorator(
        func: typing.Callable[P, typing.Any],
    ) -> typing.Callable[P, typing.Any]:
        """
        Inner decorator function to wrap the original function.

        Args:
            func (typing.Callable[P, typing.Any]): The original function to be decorated.

        Returns:
            typing.Callable[P, typing.Any]: The wrapped function.
        """
        sig = inspect.signature(func)
        __type__ = None
        for idx, parameter in enumerate(sig.parameters.values()):  # noqa B007
            if parameter.name in {"request", "websocket"}:
                __type__ = parameter.name
                break

        if __type__ is None:
            raise Exception(f'No "request" or "websocket" argument on function "{func}"')

        if __type__ == "websocket":

            @functools.wraps(func)
            async def websocket_wrapper(*args: P.args, **kwargs: P.kwargs) -> None:
                """
                Wrapper for websocket functions to enforce scope checks.

                Args:
                    *args: Positional arguments for the original function.
                    **kwargs: Keyword arguments for the original function.
                """
                websocket = kwargs.get("websocket", args[idx] if idx < len(args) else None)
                assert isinstance(websocket, WebSocket)

                if not has_required_scope(websocket, scopes_list):
                    await websocket.close()
                else:
                    await func(*args, **kwargs)

            return websocket_wrapper

        elif is_async_callable(func):

            @functools.wraps(func)
            async def async_wrapper(*args: P.args, **kwargs: P.kwargs) -> typing.Any:
                """
                Wrapper for async request/response functions to enforce scope checks.

                Args:
                    *args: Positional arguments for the original function.
                    **kwargs: Keyword arguments for the original function.

                Returns:
                    typing.Any: The result of the original function or a redirect/exception.
                """
                request = kwargs.get("request", args[idx] if idx < len(args) else None)
                assert isinstance(request, Request)

                if not has_required_scope(request, scopes_list):
                    if redirect is not None:
                        orig_request_qparam = urlencode({"next": str(request.url)})
                        next_url = f"{request.path_for(redirect)}?{orig_request_qparam}"
                        return RedirectResponse(url=next_url, status_code=303)
                    raise HTTPException(status_code=status_code)
                return await func(*args, **kwargs)

            return async_wrapper

        else:

            @functools.wraps(func)
            def sync_wrapper(*args: P.args, **kwargs: P.kwargs) -> typing.Any:
                """
                Wrapper for sync request/response functions to enforce scope checks.

                Args:
                    *args: Positional arguments for the original function.
                    **kwargs: Keyword arguments for the original function.

                Returns:
                    typing.Any: The result of the original function or a redirect/exception.
                """
                request = kwargs.get("request", args[idx] if idx < len(args) else None)
                assert isinstance(request, Request)

                if not has_required_scope(request, scopes_list):
                    if redirect is not None:
                        orig_request_qparam = urlencode({"next": str(request.url)})
                        next_url = f"{request.path_for(redirect)}?{orig_request_qparam}"
                        return RedirectResponse(url=next_url, status_code=303)
                    raise HTTPException(status_code=status_code)
                return func(*args, **kwargs)

            return sync_wrapper

    return decorator


class AuthenticationError(Exception): ...


class AuthenticationBackend(ABC):
    @abstractmethod
    async def authenticate(self, connection: Connection) -> AuthResult | None: ...


class AuthCredentials:
    def __init__(self, scopes: typing.Sequence[str] | None = None):
        self.scopes = [] if scopes is None else list(scopes)


@typing.runtime_checkable
class UserInterface(typing.Protocol):
    @property
    def is_authenticated(self) -> bool:
        raise NotImplementedError()

    @property
    def display_name(self) -> str:
        raise NotImplementedError()


# bw compatibility alias
BaseUser = UserInterface


class BasicUser(UserInterface):
    def __init__(self, username: str) -> None:
        self.username = username

    @property
    def is_authenticated(self) -> bool:
        return True

    @property
    def display_name(self) -> str:
        return self.username


class AnonymousUser(UserInterface):
    @property
    def is_authenticated(self) -> bool:
        return False

    @property
    def display_name(self) -> str:
        return ""



from __future__ import annotations

import sys
from collections.abc import Awaitable, Callable, Mapping, Sequence
from functools import cached_property
from typing import Annotated, Any, ClassVar, cast

from lilya._internal._module_loading import import_string
from lilya._utils import is_class_and_subclass
from lilya.conf import __lazy_settings__, settings as lilya_settings
from lilya.conf.exceptions import FieldException
from lilya.conf.global_settings import Settings
from lilya.datastructures import State, URLPath
from lilya.middleware.app_settings import ApplicationSettingsMiddleware
from lilya.middleware.base import DefineMiddleware
from lilya.middleware.exceptions import ExceptionMiddleware
from lilya.middleware.global_context import GlobalContextMiddleware
from lilya.middleware.server_error import ServerErrorMiddleware
from lilya.permissions.base import DefinePermission
from lilya.protocols.middleware import MiddlewareProtocol
from lilya.protocols.permissions import PermissionProtocol
from lilya.requests import Request
from lilya.responses import Response
from lilya.routing import BasePath, Include, Router
from lilya.types import (
    ApplicationType,
    ASGIApp,
    CallableDecorator,
    Doc,
    ExceptionHandler,
    Lifespan,
    Receive,
    Scope,
    Send,
)
from lilya.websockets import WebSocket

if sys.version_info >= (3, 10):  # pragma: no cover
    from typing import ParamSpec
else:  # pragma: no cover
    from typing_extensions import ParamSpec

P = ParamSpec("P")


class BaseLilya:
    router_class: ClassVar[type[Router] | None] = Router
    router: Router

    def __init__(
        self,
        debug: Annotated[bool, Doc("Enable or disable debug mode. Defaults to False.")] = False,
        settings_module: Annotated[
            Settings | str | None,
            Doc(
                """
                Alternative settings parameter. This parameter is an alternative to
                `LILYA_SETTINGS_MODULE` way of loading your settings into a Lilya application.

                When the `settings_module` is provided, it will make sure it takes priority over
                any other settings provided for the instance.

                Read more about the [settings module](https://lilya.dev/settings/)
                and how you can leverage it in your application.

                !!! Tip
                    The settings module can be very useful if you want to have, for example, a
                    [ChildLilya](https://lilya.dev/routing/#childlilya-application) that needs completely different settings
                    from the main app.

                    Example: A `ChildLilya` that takes care of the authentication into a cloud
                    provider such as AWS and handles the `boto3` module.
                """
            ),
        ] = None,
        routes: Annotated[
            Sequence[Any] | None,
            Doc(
                """
                A global `list` of lilya routes. Those routes may vary and those can
                be `Path`, `WebSocketPath` or even `Include`.

                This is also the entry-point for the routes of the application itself.

                Read more about how to use and leverage
                the [routing system](https://lilya.dev/routing/).

                **Example**

                ```python
                from lilya.apps import Lilya, Request
                from lilya.requests import Request
                from lilya.routing import Path


                async def homepage(request: Request) -> str:
                    return "Hello, home!"

                async def another(request: Request) -> str:
                    return "Hello, another!"


                app = Lilya(
                    routes=[
                        Path(
                            "/", handler=homepage,
                        ),
                        Include("/nested", routes=[
                            Path("/another", another)
                        ])
                    ]
                )
                ```

                !!! Note
                    The routing system is very powerful and this example
                    is not enough to understand what more things you can do.
                    Read in [more detail](https://lilya.dev/routing/) about this.
                """
            ),
        ] = None,
        middleware: Annotated[
            Sequence[DefineMiddleware] | None,
            Doc(
                """
                A global sequence of Lilya middlewares that are
                used by the application.

                Read more about the [Middleware](https://lilya.dev/middleware/).

                **All middlewares must be wrapped inside the `DefineMiddleware`**

                ```python
                from lilya.middleware import DefineMiddleware
                ```

                **Example**

                ```python
                from lilya.apps import Lilya
                from lilya.middleware import DefineMiddleware
                from lilya.middleware.httpsredirect import HTTPSRedirectMiddleware
                from lilya.middleware.trustedhost import TrustedHostMiddleware

                app = Lilya(
                    routes=[...],
                    middleware=[
                        DefineMiddleware(
                            TrustedHostMiddleware,
                            allowed_hosts=["example.com", "*.example.com"],
                        ),
                        DefineMiddleware(HTTPSRedirectMiddleware),
                    ],
                )
                ```
                """
            ),
        ] = None,
        exception_handlers: Annotated[
            Mapping[Any, ExceptionHandler] | None,
            Doc(
                """
                A global dictionary with handlers for exceptions.

                Read more about the [Exception handlers](https://lilya.dev/exceptions/).

                **Example**

                ```python
                from json import loads

                from lilya import status
                from lilya.apps import Lilya
                from lilya.requests import Request
                from lilya.responses import JSONResponse
                from lilya.routing import Include, Path


                async def handle_type_error(request: Request, exc: TypeError):
                    status_code = status.HTTP_400_BAD_REQUEST
                    details = loads(exc.json()) if hasattr(exc, "json") else exc.args[0]
                    return JSONResponse({"detail": details}, status_code=status_code)


                async def handle_value_error(request: Request, exc: ValueError):
                    status_code = status.HTTP_400_BAD_REQUEST
                    details = loads(exc.json()) if hasattr(exc, "json") else exc.args[0]
                    return JSONResponse({"detail": details}, status_code=status_code)


                async def me():
                    return "Hello, world!"


                app = Lilya(
                    routes=[
                        Include(
                            "/",
                            routes=[
                                Path(
                                    "/me",
                                    handler=me,
                                )
                            ],
                        )
                    ],
                    exception_handlers={
                        TypeError: handle_type_error,
                        ValueError: handle_value_error,
                    },
                )
                ```
                """
            ),
        ] = None,
        permissions: Annotated[
            Sequence[DefinePermission] | None,
            Doc(
                """
                A global sequence of Lilya permissions that are
                used by the application.

                Read more about the [Permissions](https://lilya.dev/permissions/).

                **All permissions must be wrapped inside the `DefinePermission`**

                ```python
                from lilya.permissions import DefinePermission
                ```

                **Example**

                ```python
                from lilya.apps import Lilya
                from lilya.exceptions import PermissionDenied
                from lilya.permissions import DefinePermission
                from lilya.protocols.permissions import PermissionProtocol
                from lilya.requests import Request
                from lilya.responses import Ok
                from lilya.routing import Path
                from lilya.types import ASGIApp, Receive, Scope, Send


                class AllowAccess(PermissionProtocol):
                    def __init__(self, app: ASGIApp, *args, **kwargs):
                        super().__init__(app, *args, **kwargs)
                        self.app = app

                    async def __call__(self, scope: Scope, receive: Receive, send: Send):
                        request = Request(scope=scope, receive=receive, send=send)

                        if "allow-admin" in request.headers:
                            await self.app(scope, receive, send)
                            return
                        raise PermissionDenied()


                def user(user: str):
                    return Ok({"message": f"Welcome {user}"})


                app = Lilya(
                    routes=[Path("/{user}", user)],
                    permissions=[DefinePermission(AllowAccess)],
                )
                ```
                """
            ),
        ] = None,
        on_startup: Annotated[
            Sequence[Callable[[], Any]] | None,
            Doc(
                """
                A `list` of events that are trigger upon the application
                starts.

                Read more about the [events](https://lilya.dev.dev/lifespan/).

                **Example**

                ```python
                from saffier import Database, Registry

                from lilya.apps import Lilya
                from lilya.requests import Request
                from lilya.routing import Path

                database = Database("postgresql+asyncpg://user:password@host:port/database")
                registry = Registry(database=database)


                async def create_user(request: Request):
                    # Logic to create the user
                    data = await request.json()
                    ...


                app = Lilya(
                    routes=[Path("/create", handler=create_user)],
                    on_startup=[database.connect],
                )
                ```
                """
            ),
        ] = None,
        on_shutdown: Annotated[
            Sequence[Callable[[], Any]] | None,
            Doc(
                """
                A `list` of events that are trigger upon the application
                shuts down.

                Read more about the [events](https://lilya.dev/lifespan/).

                **Example**

                ```python
                from saffier import Database, Registry

                from lilya.apps import Lilya
                from lilya.requests import Request
                from lilya.routing import Path

                database = Database("postgresql+asyncpg://user:password@host:port/database")
                registry = Registry(database=database)


                async def create_user(request: Request):
                    # Logic to create the user
                    data = await request.json()
                    ...


                app = Lilya(
                    routes=[Path("/create", handler=create_user)],
                    on_shutdown=[database.disconnect],
                )
                ```
                """
            ),
        ] = None,
        redirect_slashes: Annotated[
            bool,
            Doc(
                """
                Enable or disable automatic trailing slash redirection for HTTP routes.
                """
            ),
        ] = True,
        lifespan: Annotated[
            Lifespan[ApplicationType] | None,
            Doc(
                """
                A `lifespan` context manager handler. This is an alternative
                to `on_startup` and `on_shutdown` and you **cannot used all combined**.

                Read more about the [lifespan](https://lilya.dev/lifespan/).
                """
            ),
        ] = None,
        include_in_schema: Annotated[
            bool,
            Doc(
                """
                Enable or disable inclusion of the application in a OpenAPI schema integration.

                !!! Note
                    Lilya is not tight to any OpenAPI provider but provides some out of the box
                    flags that can be used for any possible integration.
                """
            ),
        ] = True,
    ) -> None:
        self.settings_module: Settings = None

        if settings_module is not None and isinstance(settings_module, str):
            settings_module = import_string(settings_module)

        if settings_module:
            if not isinstance(settings_module, Settings) and not is_class_and_subclass(
                settings_module, Settings
            ):
                raise FieldException("'settings_module' must be a subclass of Settings")
            elif isinstance(settings_module, Settings):
                self.settings_module = settings_module
            elif is_class_and_subclass(settings_module, Settings):
                self.settings_module = settings_module()  # type: ignore

        self.debug = self.__load_settings_value("debug", debug, is_boolean=True)

        self.exception_handlers = {} if exception_handlers is None else dict(exception_handlers)
        self.custom_middleware = self.__load_settings_value("middleware", middleware)
        self.custom_permissions = self.__load_settings_value("permissions", permissions)

        self.state = State()
        self.middleware_stack: ASGIApp | None = None

        if self.router_class is not None:
            self.router = self.router_class(
                routes=routes,
                redirect_slashes=redirect_slashes,
                permissions=self.custom_permissions,
                on_startup=on_startup,
                on_shutdown=on_shutdown,
                lifespan=lifespan,
                include_in_schema=include_in_schema,
                settings_module=self.settings,
            )
        self.__set_settings_app(self.settings, self)

    @property
    def routes(self) -> list[BasePath]:
        return self.router.routes

    def __set_settings_app(self, settings_module: Settings, app: ASGIApp) -> None:
        if settings_module.app is None:
            self.router._set_settings_app(settings_module, app)

    def __load_settings_value(
        self, name: str, value: Any | None = None, is_boolean: bool = False
    ) -> Any:
        """
        Loader used to get the settings defaults and custom settings
        of the application.
        """
        if not is_boolean:
            if not value:
                return self.__get_settings_value(
                    self.settings_module, cast(Settings, lilya_settings), name
                )
            return value

        if value is not None:
            return value
        return self.__get_settings_value(
            self.settings_module, cast(Settings, lilya_settings), name
        )

    def __get_settings_value(
        self,
        local_settings: Settings | None,
        global_settings: Settings,
        value: str,
    ) -> Any:
        """Obtains the value from a settings module or defaults to the global settings"""
        setting_value = None

        if local_settings:
            setting_value = getattr(local_settings, value, None)
        if setting_value is None:
            return getattr(global_settings, value, None)
        return setting_value

    @cached_property
    def settings(self) -> Settings:
        """
        Returns the Lilya settings object for easy access.

        This `settings` are the ones being used by the application upon
        initialisation.

        **Example**

        ```python
        from lilya.apps import Lilya

        app = Lilya()
        app.settings
        ```
        """
        general_settings = self.settings_module if self.settings_module else lilya_settings
        return cast(Settings, general_settings)

    def path_for(self, name: str, /, **path_params: Any) -> URLPath:
        return self.router.path_for(name, **path_params)

    def build_middleware_stack(self) -> ASGIApp:
        """
        Build the optimized middleware stack for the Lilya application.

        Returns:
            ASGIApp: The ASGI application with the middleware stack.
        """
        error_handler = self._get_error_handler()
        exception_handlers = self._get_exception_handlers()

        middleware = [
            DefineMiddleware(ServerErrorMiddleware, handler=error_handler, debug=self.debug),
            DefineMiddleware(GlobalContextMiddleware),
            *self.custom_middleware,
            DefineMiddleware(ApplicationSettingsMiddleware),
            DefineMiddleware(ExceptionMiddleware, handlers=exception_handlers, debug=self.debug),
        ]

        app = self.router
        for middleware_class, args, options in reversed(middleware):
            app = middleware_class(app=app, *args, **options)

        return app

    def _get_error_handler(self) -> Callable[[Request, Exception], Response] | None:
        """
        Get the error handler for middleware based on the exception handlers.

        Returns:
            Optional[Callable[[Request, Exception], Response]]: The error handler function.
        """
        return self.exception_handlers.get(500) or self.exception_handlers.get(Exception)  # type: ignore

    def _get_exception_handlers(self) -> dict[Exception, ExceptionHandler]:
        """
        Get the exception handlers for middleware based on the application's exception handlers.

        Returns:
            Dict: The exception handlers.
        """
        return {
            key: value
            for key, value in self.exception_handlers.items()
            if key not in (500, Exception)
        }

    def on_event(self, event_type: str) -> Callable:
        return self.router.on_event(event_type)

    def include(
        self,
        path: str,
        app: ASGIApp,
        name: str | None = None,
        middleware: Sequence[DefineMiddleware] | None = None,
        permissions: Sequence[DefinePermission] | None = None,
        exception_handlers: Mapping[Any, ExceptionHandler] | None = None,
        namespace: str | None = None,
        pattern: str | None = None,
        include_in_schema: bool = True,
    ) -> None:
        """
        Adds an Include application into the routes.
        """
        self.router.include(
            path=path,
            app=app,
            name=name,
            middleware=middleware,
            permissions=permissions,
            namespace=namespace,
            pattern=pattern,
            exception_handlers=exception_handlers,
            include_in_schema=include_in_schema,
        )

    def host(self, host: str, app: ASGIApp, name: str | None = None) -> None:
        """
        Adds a Host application into the routes.
        """
        self.router.host(host=host, app=app, name=name)

    def add_route(
        self,
        path: Annotated[
            str,
            Doc(
                """
                Relative path of the `Path`.
                The path can contain parameters in a dictionary like format.
                """
            ),
        ],
        handler: Annotated[
            Callable[[Request], Awaitable[Response] | Response],
            Doc(
                """
                A python callable.
                """
            ),
        ],
        methods: Annotated[
            list[str] | None,
            Doc(
                """
                List of HTTP verbs (GET, POST, PUT, DELETE, HEAD...) allowed in
                the route.
                """
            ),
        ] = None,
        name: Annotated[
            str | None,
            Doc(
                """
                The name for the PAth. The name can be reversed by `path_for()` or `reverse()`.
                """
            ),
        ] = None,
        middleware: Annotated[
            Sequence[DefineMiddleware] | None,
            Doc(
                """
                A list of middleware to run for every request.
                """
            ),
        ] = None,
        permissions: Annotated[
            Sequence[DefinePermission] | None,
            Doc(
                """
                A list of [permissions](https://lilya.dev/permissions/) to serve the application incoming requests (HTTP and Websockets).
                """
            ),
        ] = None,
        exception_handlers: Annotated[
            Mapping[Any, ExceptionHandler] | None,
            Doc(
                """
                A dictionary of [exception types](https://lilya.dev/exceptions/) (or custom exceptions) and the handler functions on an application top level. Exception handler callables should be of the form of `handler(request, exc) -> response` and may be be either standard functions, or async functions.
                """
            ),
        ] = None,
        include_in_schema: Annotated[
            bool,
            Doc(
                """
                Boolean flag indicating if it should be added to the OpenAPI docs.
                """
            ),
        ] = True,
    ) -> None:
        """
        Adds a [Path](https://lilya.dev/routing/)
        to the application routing.

        This is a dynamic way of adding routes on the fly.

        **Example**

        ```python
        from lilya.apps import Lilya


        async def hello():
            return "Hello, World!"


        app = Lilya()
        app.add_route(path="/hello", handler=hello)
        ```
        """
        self.router.add_route(
            path=path,
            handler=handler,
            methods=methods,
            name=name,
            middleware=middleware,
            permissions=permissions,
            exception_handlers=exception_handlers,
            include_in_schema=include_in_schema,
        )

    def add_websocket_route(
        self,
        path: Annotated[
            str,
            Doc(
                """
                Relative path of the `Path`.
                The path can contain parameters in a dictionary like format.
                """
            ),
        ],
        handler: Annotated[
            Callable[[WebSocket], Awaitable[None]],
            Doc(
                """
                A python callable.
                """
            ),
        ],
        name: Annotated[
            str | None,
            Doc(
                """
                The name for the WebSocketPath. The name can be reversed by `path_for()` or `reverse()`.
                """
            ),
        ] = None,
        middleware: Annotated[
            Sequence[DefineMiddleware] | None,
            Doc(
                """
                A list of middleware to run for every request.
                """
            ),
        ] = None,
        permissions: Annotated[
            Sequence[DefinePermission] | None,
            Doc(
                """
                A list of [permissions](https://lilya.dev/permissions/) to serve the application incoming requests (HTTP and Websockets).
                """
            ),
        ] = None,
        exception_handlers: Annotated[
            Mapping[Any, ExceptionHandler] | None,
            Doc(
                """
                A dictionary of [exception types](https://lilya.dev/exceptions/) (or custom exceptions) and the handler functions on an application top level. Exception handler callables should be of the form of `handler(request, exc) -> response` and may be be either standard functions, or async functions.
                """
            ),
        ] = None,
    ) -> None:
        """
        Adds a websocket [WebsocketPath](https://lilya.dev/routing/)
        to the application routing.

        This is a dynamic way of adding routes on the fly.

        **Example**

        ```python
        from lilya.apps import Lilya


        async def websocket_route(websocket):
            await websocket.accept()
            data = await websocket.receive_json()

            assert data
            await websocket.send_json({"data": "lilya"})
            await websocket.close()


        app = Lilya()
        app.add_websocket_route(path="/ws", handler=websocket_route)
        ```
        """
        self.router.add_websocket_route(
            path=path,
            handler=handler,
            name=name,
            middleware=middleware,
            permissions=permissions,
            exception_handlers=exception_handlers,
        )

    def add_middleware(
        self, middleware: type[MiddlewareProtocol], *args: P.args, **kwargs: P.kwargs
    ) -> None:
        """
        Adds an external middleware to the stack.
        """
        if self.middleware_stack is not None:
            raise RuntimeError("Middlewares cannot be added once the application has started.")
        self.custom_middleware.insert(0, DefineMiddleware(middleware, *args, **kwargs))

    def add_permission(
        self, permission: type[PermissionProtocol], *args: P.args, **kwargs: P.kwargs
    ) -> None:
        """
        Adds an external permissions to the stack.
        """
        if self.router.permission_started:
            raise RuntimeError("Permissions cannot be added once the application has started.")
        self.router.permissions.insert(0, DefinePermission(permission, *args, **kwargs))

    def add_exception_handler(
        self,
        exception_cls_or_status_code: int | type[Exception],
        handler: ExceptionHandler,
    ) -> None:
        self.exception_handlers[exception_cls_or_status_code] = handler

    def add_event_handler(self, event_type: str, func: Callable[[], Any]) -> None:
        self.router.add_event_handler(event_type, func)

    def add_child_lilya(
        self,
        path: str,
        child: Annotated[
            ChildLilya,
            Doc(
                """
                The [ChildLilya](https://lilya.dev/routing/#childlilya-application) instance
                to be added.
                """
            ),
        ],
        name: str | None = None,
        middleware: Sequence[DefineMiddleware] | None = None,
        permissions: Sequence[DefinePermission] | None = None,
        exception_handlers: Mapping[Any, ExceptionHandler] | None = None,
        include_in_schema: bool | None = True,
        deprecated: bool | None = None,
    ) -> None:
        """
        Adds a [ChildLilya](https://lilya.dev/routing/#childlilya-application) directly to the active application router.

        **Example**

        ```python
        from lilya.apps import ChildLilya, Lilya
        from lilya.routing import Path, Include


        async def hello(self):
            return "Hello, World!"

        child = ChildLilya(routes=[Path("/", handler=hello)])

        app = Lilya()
        app.add_child_lilya(path"/child", child=child)
        ```
        """
        if not isinstance(child, ChildLilya):
            raise ValueError("The child must be an instance of a ChildLilya.")

        self.router.routes.append(
            Include(
                path=path,
                name=name,
                app=child,
                middleware=middleware,
                permissions=permissions,
                exception_handlers=exception_handlers,
                include_in_schema=include_in_schema,
                deprecated=deprecated,
            )
        )

    async def _globalise_settings(self) -> None:
        """
        Making sure the global settings remain as is
        after the request is done.
        """
        lilya_settings.configure(__lazy_settings__._wrapped)

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        scope["app"] = self

        if self.middleware_stack is None:
            self.middleware_stack = self.build_middleware_stack()
        await self.middleware_stack(scope, receive, send)
        await self._globalise_settings()


class Lilya(BaseLilya):
    """
    Initialize the Lilya ASGI framework.

    !!! Tip
        All the parameters available in the object have defaults being loaded by the
        [settings system](https://lilya.dev/settings/) if nothing is provided.

    **Example**:

    ```python
    from lilya import Lilya

    app = Lilya(debug=True, routes=[...], middleware=[...], ...)
    ```
    """

    def get(
        self,
        path: str,
        name: str | None = None,
        middleware: Sequence[DefineMiddleware] | None = None,
        permissions: Sequence[DefinePermission] | None = None,
        exception_handlers: Mapping[Any, ExceptionHandler] | None = None,
        include_in_schema: bool = True,
    ) -> Callable[[CallableDecorator], CallableDecorator]:
        """
        Decorator for defining a GET route.

        Args:
            path (str): The URL path pattern for the route.
            methods (list[str] | None, optional): The HTTP methods allowed for the route. Defaults to None.
            name (str | None, optional): The name of the route. Defaults to None.
            middleware (Sequence[DefineMiddleware] | None, optional): The middleware functions to apply to the route. Defaults to None.
            permissions (Sequence[DefinePermission] | None, optional): The permissions required for the route. Defaults to None.
            exception_handlers (Mapping[Any, ExceptionHandler] | None, optional): The exception handlers for the route. Defaults to None.
            include_in_schema (bool, optional): Whether to include the route in the API schema. Defaults to True.

        Returns:
            Callable[[CallableDecorator], CallableDecorator]: The decorated function.
        """

        return self.router.get(
            path=path,
            name=name,
            middleware=middleware,
            permissions=permissions,
            exception_handlers=exception_handlers,
            include_in_schema=include_in_schema,
        )

    def head(
        self,
        path: str,
        name: str | None = None,
        middleware: Sequence[DefineMiddleware] | None = None,
        permissions: Sequence[DefinePermission] | None = None,
        exception_handlers: Mapping[Any, ExceptionHandler] | None = None,
        include_in_schema: bool = True,
    ) -> Callable[[CallableDecorator], CallableDecorator]:
        """
        Decorator for defining a HEAD route.

        Args:
            path (str): The URL path pattern for the route.
            methods (list[str] | None, optional): The HTTP methods allowed for the route. Defaults to None.
            name (str | None, optional): The name of the route. Defaults to None.
            middleware (Sequence[DefineMiddleware] | None, optional): The middleware functions to apply to the route. Defaults to None.
            permissions (Sequence[DefinePermission] | None, optional): The permissions required for the route. Defaults to None.
            exception_handlers (Mapping[Any, ExceptionHandler] | None, optional): The exception handlers for the route. Defaults to None.
            include_in_schema (bool, optional): Whether to include the route in the API schema. Defaults to True.

        Returns:
            Callable[[CallableDecorator], CallableDecorator]: The decorated function.
        """

        return self.router.head(
            path=path,
            name=name,
            middleware=middleware,
            permissions=permissions,
            exception_handlers=exception_handlers,
            include_in_schema=include_in_schema,
        )

    def post(
        self,
        path: str,
        name: str | None = None,
        middleware: Sequence[DefineMiddleware] | None = None,
        permissions: Sequence[DefinePermission] | None = None,
        exception_handlers: Mapping[Any, ExceptionHandler] | None = None,
        include_in_schema: bool = True,
    ) -> Callable[[CallableDecorator], CallableDecorator]:
        """
        Decorator for defining a POST route.

        Args:
            path (str): The URL path pattern for the route.
            methods (list[str] | None, optional): The HTTP methods allowed for the route. Defaults to None.
            name (str | None, optional): The name of the route. Defaults to None.
            middleware (Sequence[DefineMiddleware] | None, optional): The middleware functions to apply to the route. Defaults to None.
            permissions (Sequence[DefinePermission] | None, optional): The permissions required for the route. Defaults to None.
            exception_handlers (Mapping[Any, ExceptionHandler] | None, optional): The exception handlers for the route. Defaults to None.
            include_in_schema (bool, optional): Whether to include the route in the API schema. Defaults to True.

        Returns:
            Callable[[CallableDecorator], CallableDecorator]: The decorated function.
        """

        return self.router.post(
            path=path,
            name=name,
            middleware=middleware,
            permissions=permissions,
            exception_handlers=exception_handlers,
            include_in_schema=include_in_schema,
        )

    def put(
        self,
        path: str,
        name: str | None = None,
        middleware: Sequence[DefineMiddleware] | None = None,
        permissions: Sequence[DefinePermission] | None = None,
        exception_handlers: Mapping[Any, ExceptionHandler] | None = None,
        include_in_schema: bool = True,
    ) -> Callable[[CallableDecorator], CallableDecorator]:
        """
        Decorator for defining a PUT route.

        Args:
            path (str): The URL path pattern for the route.
            methods (list[str] | None, optional): The HTTP methods allowed for the route. Defaults to None.
            name (str | None, optional): The name of the route. Defaults to None.
            middleware (Sequence[DefineMiddleware] | None, optional): The middleware functions to apply to the route. Defaults to None.
            permissions (Sequence[DefinePermission] | None, optional): The permissions required for the route. Defaults to None.
            exception_handlers (Mapping[Any, ExceptionHandler] | None, optional): The exception handlers for the route. Defaults to None.
            include_in_schema (bool, optional): Whether to include the route in the API schema. Defaults to True.

        Returns:
            Callable[[CallableDecorator], CallableDecorator]: The decorated function.
        """

        return self.router.put(
            path=path,
            name=name,
            middleware=middleware,
            permissions=permissions,
            exception_handlers=exception_handlers,
            include_in_schema=include_in_schema,
        )

    def patch(
        self,
        path: str,
        name: str | None = None,
        middleware: Sequence[DefineMiddleware] | None = None,
        permissions: Sequence[DefinePermission] | None = None,
        exception_handlers: Mapping[Any, ExceptionHandler] | None = None,
        include_in_schema: bool = True,
    ) -> Callable[[CallableDecorator], CallableDecorator]:
        """
        Decorator for defining a PATCH route.

        Args:
            path (str): The URL path pattern for the route.
            methods (list[str] | None, optional): The HTTP methods allowed for the route. Defaults to None.
            name (str | None, optional): The name of the route. Defaults to None.
            middleware (Sequence[DefineMiddleware] | None, optional): The middleware functions to apply to the route. Defaults to None.
            permissions (Sequence[DefinePermission] | None, optional): The permissions required for the route. Defaults to None.
            exception_handlers (Mapping[Any, ExceptionHandler] | None, optional): The exception handlers for the route. Defaults to None.
            include_in_schema (bool, optional): Whether to include the route in the API schema. Defaults to True.

        Returns:
            Callable[[CallableDecorator], CallableDecorator]: The decorated function.
        """

        return self.router.patch(
            path=path,
            name=name,
            middleware=middleware,
            permissions=permissions,
            exception_handlers=exception_handlers,
            include_in_schema=include_in_schema,
        )

    def delete(
        self,
        path: str,
        name: str | None = None,
        middleware: Sequence[DefineMiddleware] | None = None,
        permissions: Sequence[DefinePermission] | None = None,
        exception_handlers: Mapping[Any, ExceptionHandler] | None = None,
        include_in_schema: bool = True,
    ) -> Callable[[CallableDecorator], CallableDecorator]:
        """
        Decorator for defining a DELETE route.

        Args:
            path (str): The URL path pattern for the route.
            methods (list[str] | None, optional): The HTTP methods allowed for the route. Defaults to None.
            name (str | None, optional): The name of the route. Defaults to None.
            middleware (Sequence[DefineMiddleware] | None, optional): The middleware functions to apply to the route. Defaults to None.
            permissions (Sequence[DefinePermission] | None, optional): The permissions required for the route. Defaults to None.
            exception_handlers (Mapping[Any, ExceptionHandler] | None, optional): The exception handlers for the route. Defaults to None.
            include_in_schema (bool, optional): Whether to include the route in the API schema. Defaults to True.

        Returns:
            Callable[[CallableDecorator], CallableDecorator]: The decorated function.
        """

        return self.router.delete(
            path=path,
            name=name,
            middleware=middleware,
            permissions=permissions,
            exception_handlers=exception_handlers,
            include_in_schema=include_in_schema,
        )

    def trace(
        self,
        path: str,
        name: str | None = None,
        middleware: Sequence[DefineMiddleware] | None = None,
        permissions: Sequence[DefinePermission] | None = None,
        exception_handlers: Mapping[Any, ExceptionHandler] | None = None,
        include_in_schema: bool = True,
    ) -> Callable[[CallableDecorator], CallableDecorator]:
        """
        Decorator for defining a TRACE route.

        Args:
            path (str): The URL path pattern for the route.
            methods (list[str] | None, optional): The HTTP methods allowed for the route. Defaults to None.
            name (str | None, optional): The name of the route. Defaults to None.
            middleware (Sequence[DefineMiddleware] | None, optional): The middleware functions to apply to the route. Defaults to None.
            permissions (Sequence[DefinePermission] | None, optional): The permissions required for the route. Defaults to None.
            exception_handlers (Mapping[Any, ExceptionHandler] | None, optional): The exception handlers for the route. Defaults to None.
            include_in_schema (bool, optional): Whether to include the route in the API schema. Defaults to True.

        Returns:
            Callable[[CallableDecorator], CallableDecorator]: The decorated function.
        """

        return self.router.trace(
            path=path,
            name=name,
            middleware=middleware,
            permissions=permissions,
            exception_handlers=exception_handlers,
            include_in_schema=include_in_schema,
        )

    def options(
        self,
        path: str,
        name: str | None = None,
        middleware: Sequence[DefineMiddleware] | None = None,
        permissions: Sequence[DefinePermission] | None = None,
        exception_handlers: Mapping[Any, ExceptionHandler] | None = None,
        include_in_schema: bool = True,
    ) -> Callable[[CallableDecorator], CallableDecorator]:
        """
        Decorator for defining a OPTIONS route.

        Args:
            path (str): The URL path pattern for the route.
            methods (list[str] | None, optional): The HTTP methods allowed for the route. Defaults to None.
            name (str | None, optional): The name of the route. Defaults to None.
            middleware (Sequence[DefineMiddleware] | None, optional): The middleware functions to apply to the route. Defaults to None.
            permissions (Sequence[DefinePermission] | None, optional): The permissions required for the route. Defaults to None.
            exception_handlers (Mapping[Any, ExceptionHandler] | None, optional): The exception handlers for the route. Defaults to None.
            include_in_schema (bool, optional): Whether to include the route in the API schema. Defaults to True.

        Returns:
            Callable[[CallableDecorator], CallableDecorator]: The decorated function.
        """

        return self.router.options(
            path=path,
            name=name,
            middleware=middleware,
            permissions=permissions,
            exception_handlers=exception_handlers,
            include_in_schema=include_in_schema,
        )

    def route(
        self,
        path: str,
        methods: list[str],
        name: str | None = None,
        middleware: Sequence[DefineMiddleware] | None = None,
        permissions: Sequence[DefinePermission] | None = None,
        exception_handlers: Mapping[Any, ExceptionHandler] | None = None,
        include_in_schema: bool = True,
    ) -> Callable[[CallableDecorator], CallableDecorator]:
        """
        Decorator for defining a generic route.

        Args:
            path (str): The URL path pattern for the route.
            methods (list[str] | None, optional): The HTTP methods allowed for the route. Defaults to None.
            name (str | None, optional): The name of the route. Defaults to None.
            middleware (Sequence[DefineMiddleware] | None, optional): The middleware functions to apply to the route. Defaults to None.
            permissions (Sequence[DefinePermission] | None, optional): The permissions required for the route. Defaults to None.
            exception_handlers (Mapping[Any, ExceptionHandler] | None, optional): The exception handlers for the route. Defaults to None.
            include_in_schema (bool, optional): Whether to include the route in the API schema. Defaults to True.

        Returns:
            Callable[[CallableDecorator], CallableDecorator]: The decorated function.
        """

        return self.router.route(
            path=path,
            methods=methods,
            name=name,
            middleware=middleware,
            permissions=permissions,
            exception_handlers=exception_handlers,
            include_in_schema=include_in_schema,
        )

    def websocket(
        self,
        path: str,
        name: str | None = None,
        middleware: Sequence[DefineMiddleware] | None = None,
        permissions: Sequence[DefinePermission] | None = None,
        exception_handlers: Mapping[Any, ExceptionHandler] | None = None,
    ) -> Callable[[CallableDecorator], CallableDecorator]:
        """
        Decorator for defining a WebSocket route.

        Args:
            path (str): The URL path for the WebSocket route.
            name (str, optional): The name of the route. Defaults to None.
            middleware (Sequence[DefineMiddleware], optional): The middleware to apply to the route. Defaults to None.
            permissions (Sequence[DefinePermission], optional): The permissions required for the route. Defaults to None.
            exception_handlers (Mapping[Any, ExceptionHandler], optional): The exception handlers for the route. Defaults to None.

        Returns:
            Callable[[CallableDecorator], CallableDecorator]: The decorated function.

        """

        return self.router.websocket(
            path=path,
            name=name,
            middleware=middleware,
            permissions=permissions,
            exception_handlers=exception_handlers,
        )


class ChildLilya(Lilya):
    """
    `ChildLilya` application object. The main entry-point for a modular application/API
    with Lilya.

    The `ChildLilya` inherits directly from the `Lilya` object which means all the same
    parameters, attributes and functions of Lilya ara also available in the `ChildLilya`.


    !!! Tip
        All the parameters available in the object have defaults being loaded by the
        [settings system](https://lilya.dev/settings/) if nothing is provided.

    ## Example

    ```python
    from lilya.apps import Lilya, ChildLilya
    from lilya.routing import Include


    app = Lilya(routes=[Include("/child", app=ChildLilya(...))])
    ```
    """



from __future__ import annotations

import html
import inspect
import traceback
from collections.abc import Callable
from typing import Any

from lilya import status
from lilya.compat import is_async_callable
from lilya.concurrency import run_in_threadpool
from lilya.enums import MediaType, ScopeType
from lilya.middleware.styles.errors import (
    get_center_line,
    get_css_style,
    get_frame,
    get_js,
    get_line,
    get_template_errors,
)
from lilya.protocols.middleware import MiddlewareProtocol
from lilya.requests import Request
from lilya.responses import HTMLResponse, PlainText, Response
from lilya.types import ASGIApp, Message, Receive, Scope, Send


class ServerErrorMiddleware(MiddlewareProtocol):
    """
    Middleware for handling server errors and returning appropriate 500 responses.

    If 'debug' is set, traceback responses will be returned;
    otherwise, the designated 'handler' will be called.
    """

    def __init__(
        self,
        app: ASGIApp,
        handler: Callable[[Request, Exception], Any] | None = None,
        debug: bool = False,
    ) -> None:
        """
        Initialize the ServerErrorMiddleware.

        Args:
            app (ASGIApp): The ASGI application.
            handler (Optional[Callable]): Custom error handler.
            debug (bool): Enable debug mode for returning traceback responses.
        """
        self.app = app
        self.handler = handler
        self.debug = debug

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Handle incoming ASGI scope.

        Args:
            scope (Scope): ASGI scope.
            receive (Receive): ASGI receive function.
            send (Send): ASGI send function.
        """
        if scope["type"] != ScopeType.HTTP:
            await self.app(scope, receive, send)
            return

        response_started = False

        async def _send(message: Message) -> None:
            nonlocal response_started, send

            if message["type"] == "http.response.start":
                response_started = True
            await send(message)

        try:
            await self.app(scope, receive, _send)
        except Exception as exc:
            request = Request(scope)
            response = await self.handle_exception(request, exc, scope, receive, send)
            await self.send_response(response, scope, receive, send, response_started)
            raise exc

    async def handle_exception(
        self, request: Request, exc: Exception, scope: Scope, receive: Receive, send: Send
    ) -> Any:
        """
        Handle the exception and return an appropriate response.

        Args:
            request (Request): The incoming HTTP request.
            exc (Exception): The raised exception.
            scope (Scope): The ASGI scope.
            receive (Receive): ASGI receive function.
            send (Send): ASGI send function.

        Returns:
            The appropriate HTTP response.
        """
        if self.debug:
            return self.debug_response(request, exc)
        elif self.handler is None:
            return self.error_response(request, exc)
        else:
            return await self.run_handler(request, exc)

    async def run_handler(self, request: Request, exc: Exception) -> Any:
        """
        Run the custom error handler.

        Args:
            request (Request): The incoming HTTP request.
            exc (Exception): The raised exception.

        Returns:
            The response generated by the custom error handler.
        """
        if is_async_callable(self.handler):
            return await self.handler(request, exc)
        else:
            return await run_in_threadpool(self.handler, request, exc)

    async def send_response(
        self,
        response: Response,
        scope: Scope,
        receive: Receive,
        send: Send,
        response_started: bool,
    ) -> None:
        """
        Send the generated response.

        Args:
            response (Response): The HTTP response to send.
            scope (Scope): The ASGI scope.
            receive (Receive): ASGI receive function.
            send (Send): ASGI send function.
        """
        if not response_started:
            await response(scope, receive, send)

    def error_response(self, request: Request, exc: Exception) -> Response:
        """
        Generate a plain text 500 error response.

        Args:
            request (Request): The incoming HTTP request.
            exc (Exception): The raised exception.

        Returns:
            Response: The plain text 500 error response.
        """
        return PlainText(
            "Internal Server Error", status_code=status.HTTP_500_INTERNAL_SERVER_ERROR
        )

    def debug_response(self, request: Request, exc: Exception) -> Response:
        """
        Generate a debug response with HTML or plain text.

        Args:
            request (Request): The incoming HTTP request.
            exc (Exception): The raised exception.

        Returns:
            Response: The debug response.
        """
        accept = request.headers.get("accept", "")

        if MediaType.HTML in accept:
            content = self.generate_html(exc)
            return HTMLResponse(content, status_code=status.HTTP_500_INTERNAL_SERVER_ERROR)
        content = self.generate_plain_text(exc)
        return PlainText(content, status_code=status.HTTP_500_INTERNAL_SERVER_ERROR)

    def generate_html(self, exc: Exception) -> str:
        """
        Generate HTML content for the debug response.

        Args:
            exc (Exception): The raised exception.

        Returns:
            str: The HTML content.
        """
        traceback_obj = traceback.TracebackException.from_exception(exc, capture_locals=True)

        exc_html = ""
        is_collapsed = False
        exc_traceback = exc.__traceback__
        if exc_traceback is not None:
            frames = inspect.getinnerframes(exc_traceback, 7)
            for frame in reversed(frames):
                exc_html += self.generate_frame_html(frame, is_collapsed)
                is_collapsed = True

        # escape error class and text
        try:
            exc_type_str = traceback_obj.exc_type_str
        except Exception:
            # for older python versions < 3.13
            exc_type_str = traceback_obj.exc_type.__name__
        error = f"{html.escape(exc_type_str)}: " f"{html.escape(str(traceback_obj))}"

        template = get_template_errors()
        return template.format(styles=get_css_style(), js=get_js(), error=error, exc_html=exc_html)

    def generate_plain_text(self, exc: Exception) -> str:
        """
        Generate plain text content for the debug response.

        Args:
            exc (Exception): The raised exception.

        Returns:
            str: The plain text content.
        """
        return "".join(traceback.format_exception(type(exc), exc, exc.__traceback__))

    def format_line(self, index: int, line: str, frame_lineno: int, frame_index: int) -> str:
        """
        Format a single line in a traceback frame.

        Args:
            index (int): The index of the line.
            line (str): The content of the line.
            frame_lineno (int): Line number in the frame.
            frame_index (int): Index of the frame.

        Returns:
            str: The formatted line HTML.
        """
        escaped_line = html.escape(line).replace(" ", "&nbsp")
        lineno = (frame_lineno - frame_index) + index

        values = {
            "line": escaped_line,
            "lineno": lineno,
        }

        return (
            get_line().format(**values)
            if index != frame_index
            else get_center_line().format(**values)
        )

    def generate_frame_html(self, frame: inspect.FrameInfo, is_collapsed: bool) -> str:
        """
        Generate HTML content for a single frame in the traceback.

        Args:
            frame (FrameInfo): Information about the frame.
            is_collapsed (bool): Whether the frame is collapsed.

        Returns:
            str: The HTML content for the frame.
        """
        code_context = "".join(
            self.format_line(
                index,
                line,
                frame.lineno,
                frame.index,
            )
            for index, line in enumerate(frame.code_context or [])
        )

        values = {
            "frame_filename": html.escape(frame.filename),
            "frame_lineno": frame.lineno,
            "frame_name": html.escape(frame.function),
            "code_context": code_context,
            "collapsed": "collapsed" if is_collapsed else "",
            "collapse_button": "+" if is_collapsed else "&#8210;",
        }
        return get_frame().format(**values)



from __future__ import annotations

import inspect
import sys
from typing import Any, TypedDict

from lilya.compat import is_async_callable
from lilya.testclient._internal.types import ASGI2App, ASGI3App
from lilya.types import Receive, Scope, Send

if sys.version_info >= (3, 10):  # pragma: no cover
    from typing import TypeGuard
else:  # pragma: no cover
    from typing_extensions import TypeGuard


def is_asgi3(app: ASGI2App | ASGI3App) -> TypeGuard[ASGI3App]:
    if inspect.isclass(app):
        return hasattr(app, "__await__")
    return is_async_callable(app)


class WrapASGI2:
    """
    Provide an ASGI3 interface onto an ASGI2 app.
    """

    def __init__(self, app: ASGI2App) -> None:
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        instance = self.app(scope)
        await instance(receive, send)


class AsyncBackend(TypedDict):
    backend: str
    backend_options: dict[str, Any]



from __future__ import annotations

from typing import TYPE_CHECKING, Any, TypeVar, cast

from lilya.compat import is_async_callable
from lilya.types import ASGIApp, Lifespan, LifespanEvent, Receive, Scope, Send

if TYPE_CHECKING:
    from lilya.routing import BaseRouter

T = TypeVar("T")


class AsyncLifespanContextManager:  # pragma: no cover
    """
    Manages and handles the on_startup and on_shutdown events
    in a Lilya way.
    """

    def __init__(
        self,
        on_shutdown: LifespanEvent | None = None,
        on_startup: LifespanEvent | None = None,
    ) -> None:
        self.on_startup = [] if on_startup is None else list(on_startup)
        self.on_shutdown = [] if on_shutdown is None else list(on_shutdown)

    def __call__(self: T, app: ASGIApp | Any) -> T:
        return self

    async def __aenter__(self) -> None:
        """Runs the functions on startup"""
        for handler in self.on_startup:
            if is_async_callable(handler):
                await handler()
            else:
                handler()

    async def __aexit__(self, scope: Scope, receive: Receive, send: Send, **kwargs: Any) -> None:
        """Runs the functions on shutdown"""
        for handler in self.on_shutdown:
            if is_async_callable(handler):
                await handler()
            else:
                handler()


AyncLifespanContextManager = AsyncLifespanContextManager


class AsyncLifespan:
    def __init__(self, router: BaseRouter):
        self.router = router

    async def __aenter__(self) -> None:
        await self.router.startup()

    async def __aexit__(self, *exc_info: object) -> None:
        await self.router.shutdown()

    def __call__(self: T, app: object) -> T:
        return self


def handle_lifespan_events(
    on_startup: LifespanEvent | None = None,
    on_shutdown: LifespanEvent | None = None,
    lifespan: Lifespan[Any] | None = None,
) -> Lifespan[Any] | None:  # pragma: no cover
    if on_startup or on_shutdown:
        return cast(
            Lifespan[Any],
            AsyncLifespanContextManager(on_startup=on_startup, on_shutdown=on_shutdown),
        )
    elif lifespan:
        return lifespan
    return None


def generate_lifespan_events(
    on_startup: LifespanEvent | None = None,
    on_shutdown: LifespanEvent | None = None,
    lifespan: Lifespan[Any] | None = None,
) -> Lifespan[Any]:  # pragma: no cover
    if lifespan:
        return lifespan
    return cast(
        Lifespan[Any], AsyncLifespanContextManager(on_startup=on_startup, on_shutdown=on_shutdown)
    )



#!/usr/bin/env python
# type: ignore
from __future__ import annotations

import contextlib
import math
from collections.abc import Generator, MutableMapping, Sequence
from concurrent.futures import Future
from typing import Any, Literal, cast
from urllib.parse import urljoin

import anyio
import anyio.from_thread
import httpx
from anyio.abc import ObjectReceiveStream, ObjectSendStream
from anyio.streams.stapled import StapledObjectStream
from httpx._types import (
    AuthTypes,
    CookieTypes,
    HeaderTypes,
    QueryParamTypes,
    RequestContent,
    RequestFiles,
    TimeoutTypes,
    URLTypes,
)

from lilya.testclient._internal.inputs import RequestInputsDefaultValues
from lilya.testclient._internal.transport import TestClientTransport
from lilya.testclient._internal.types import ASGI2App, RequestData
from lilya.testclient._internal.utils import AsyncBackend, WrapASGI2, is_asgi3
from lilya.testclient._internal.websockets import WebSocketTestSession
from lilya.testclient.exceptions import UpgradeException
from lilya.types import ASGIApp

try:
    import httpx

except ModuleNotFoundError:  # pragma: no cover
    raise RuntimeError(
        "The lilya.testclient module requires the httpx package to be installed.\n"
        "You can install this with:\n"
        "    $ pip install httpx\n"
    ) from None


class TestClient(httpx.Client):
    """
    A test client for making HTTP requests to an ASGI application.

    Args:
        app (ASGIApp): The ASGI application to test.
        base_url (str, optional): The base URL for the requests. Defaults to "http://testserver".
        raise_server_exceptions (bool, optional): Whether to raise exceptions for server errors. Defaults to True.
        root_path (str, optional): The root path for the requests. Defaults to "".
        backend (Literal["asyncio", "trio"], optional): The async backend to use. Defaults to "asyncio".
        backend_options (dict[str, Any] | None, optional): Options for the async backend. Defaults to None.
        cookies (httpx._types.CookieTypes | None, optional): Cookies to include in the requests. Defaults to None.
        headers (dict[str, str] | None, optional): Headers to include in the requests. Defaults to None.
        follow_redirects (bool, optional): Whether to follow redirects. Defaults to True.
        check_asgi_conformance (bool, optional): Whether to check the asgi conformance. Defaults to True.
    """

    __test__ = False
    task: Future[None]
    portal: anyio.abc.BlockingPortal | None = None

    def __init__(
        self,
        app: ASGIApp,
        base_url: str = "http://testserver",
        raise_server_exceptions: bool = True,
        root_path: str = "",
        backend: Literal["asyncio", "trio"] = "asyncio",
        backend_options: dict[str, Any] | None = None,
        cookies: CookieTypes | None = None,
        headers: HeaderTypes | None = None,
        follow_redirects: bool = True,
        check_asgi_conformance: bool = True,
    ) -> None:
        self.async_backend = AsyncBackend(backend=backend, backend_options=backend_options or {})
        if is_asgi3(app):
            asgi_app = app
        else:
            app = cast(ASGI2App, app)  # type: ignore[assignment]
            asgi_app = WrapASGI2(app)  # type: ignore[arg-type]
        self.app = asgi_app
        self.app_state: dict[str, Any] = {}
        transport = TestClientTransport(
            app=self.app,
            portal_factory=self._portal_factory,
            raise_server_exceptions=raise_server_exceptions,
            root_path=root_path,
            app_state=self.app_state,
            check_asgi_conformance=check_asgi_conformance,
        )
        if headers is None:
            headers = {}
        headers.setdefault("user-agent", "testclient")
        super().__init__(
            base_url=base_url,
            headers=headers,
            transport=transport,
            follow_redirects=follow_redirects,
            cookies=cookies,
        )

    @contextlib.contextmanager
    def _portal_factory(self) -> Generator[anyio.abc.BlockingPortal, None, None]:
        """
        Context manager that creates a blocking portal for handling async tasks.

        Yields:
            anyio.abc.BlockingPortal: The blocking portal.
        """
        if self.portal is not None:
            yield self.portal
        else:
            with anyio.from_thread.start_blocking_portal(**self.async_backend) as portal:
                yield portal

    def request(
        self,
        method: str,
        url: URLTypes,
        *,
        content: RequestContent | None = None,
        data: RequestData | None = None,
        files: RequestFiles | None = None,
        json: Any = None,
        params: QueryParamTypes | None = None,
        headers: HeaderTypes | None = None,
        cookies: CookieTypes | None = None,
        auth: AuthTypes | httpx._client.UseClientDefault = httpx._client.USE_CLIENT_DEFAULT,
        follow_redirects: bool | None = None,
        timeout: TimeoutTypes | httpx._client.UseClientDefault = httpx._client.USE_CLIENT_DEFAULT,
        extensions: dict[str, Any] | None = None,
    ) -> httpx.Response:
        """
        Sends an HTTP request.

        Args:
            method (str): The HTTP method.
            url (URLTypes): The URL to send the request to.
            content (RequestContent | None, optional): The request content. Defaults to None.
            data (RequestData | None, optional): The request data. Defaults to None.
            files (RequestFiles | None, optional): The request files. Defaults to None.
            json (Any, optional): The request JSON. Defaults to None.
            params (QueryParamTypes | None, optional): The request query parameters. Defaults to None.
            headers (HeaderTypes | None, optional): The request headers. Defaults to None.
            cookies (CookieTypes | None, optional): The request cookies. Defaults to None.
            auth (AuthTypes | httpx._client.UseClientDefault, optional): The request authentication. Defaults to httpx._client.USE_CLIENT_DEFAULT.
            follow_redirects (bool | None, optional): Whether to follow redirects. Defaults to None.
            timeout (TimeoutTypes | httpx._client.UseClientDefault, optional): The request timeout. Defaults to httpx._client.USE_CLIENT_DEFAULT.
            extensions (dict[str, Any] | None, optional): The request extensions. Defaults to None.

        Returns:
            httpx.Response: The HTTP response.
        """
        url = self._merge_url(url)
        redirect: bool | httpx._client.UseClientDefault = httpx._client.USE_CLIENT_DEFAULT
        if follow_redirects is not None:
            redirect = follow_redirects

        return super().request(
            method,
            url,
            content=content,
            data=data,
            files=files,
            json=json,
            params=params,
            headers=headers,
            cookies=cookies,
            auth=auth,
            follow_redirects=redirect,
            timeout=timeout,
            extensions=extensions,
        )

    def _process_request(self, method: str, url: URLTypes, **kwargs: Any) -> httpx.Request:
        """
        Processes the request.

        Args:
            method (str): The string method to request from.
            url (URLTypes): The URL to send the request to.
            **kwargs (RequestInputs): The request inputs.

        Returns:
            httpx.Request: The HTTP request.
        """
        if not kwargs:
            kwargs = RequestInputsDefaultValues  # type: ignore
        else:
            remaining_kwargs = {
                k: v for k, v in RequestInputsDefaultValues.items() if k not in kwargs
            }
            kwargs.update(remaining_kwargs)  # type: ignore

        return self.request(method=method, url=url, **kwargs)  # type: ignore

    def get(
        self,
        url: URLTypes,
        **kwargs: Any,
    ) -> httpx.Response:
        return self._process_request(method="GET", url=url, **kwargs)  # type: ignore

    def head(
        self,
        url: URLTypes,
        **kwargs: Any,
    ) -> httpx.Response:
        return self._process_request(method="HEAD", url=url, **kwargs)  # type: ignore

    def post(
        self,
        url: URLTypes,
        **kwargs: Any,
    ) -> httpx.Response:
        return self._process_request(method="POST", url=url, **kwargs)  # type: ignore

    def put(
        self,
        url: URLTypes,
        **kwargs: Any,
    ) -> httpx.Response:
        return self._process_request(method="PUT", url=url, **kwargs)  # type: ignore

    def patch(
        self,
        url: URLTypes,
        **kwargs: Any,
    ) -> httpx.Response:
        return self._process_request(method="PATCH", url=url, **kwargs)  # type: ignore

    def delete(
        self,
        url: URLTypes,
        **kwargs: Any,
    ) -> httpx.Response:
        return self._process_request(method="DELETE", url=url, **kwargs)  # type: ignore

    def options(
        self,
        url: URLTypes,
        **kwargs: Any,
    ) -> httpx.Response:
        return self._process_request(method="OPTIONS", url=url, **kwargs)  # type: ignore

    def websocket_connect(
        self,
        url: str,
        subprotocols: Sequence[str] | None = None,
        **kwargs: Any,
    ) -> WebSocketTestSession:
        """
        Establishes a WebSocket connection.

        Args:
            url (str): The WebSocket URL.
            subprotocols (Sequence[str] | None, optional): The WebSocket subprotocols. Defaults to None.
            **kwargs (Any): Additional keyword arguments.

        Returns:
            WebSocketTestSession: The WebSocket session.
        """
        url = urljoin("ws://testserver", url)
        headers = self._prepare_websocket_headers(subprotocols, **kwargs)
        kwargs["headers"] = headers
        try:
            super().request("GET", url, **kwargs)
        except UpgradeException as exc:
            session = exc.session
        else:
            raise RuntimeError("Expected WebSocket upgrade")

        return session

    def _prepare_websocket_headers(
        self,
        subprotocols: Sequence[str] | None = None,
        **kwargs: dict[str, Any],
    ) -> dict[str, str]:
        """
        Prepare the headers for a WebSocket connection.

        Args:
            subprotocols (Sequence[str] | None, optional): The WebSocket subprotocols. Defaults to None.
            **kwargs (dict[str, Any]): Additional keyword arguments.

        Returns:
            dict[str, str]: The prepared headers.
        """
        headers = kwargs.get("headers", {})
        headers.setdefault("connection", "upgrade")
        headers.setdefault("sec-websocket-key", "testserver==")
        headers.setdefault("sec-websocket-version", "13")
        if subprotocols is not None:
            headers.setdefault("sec-websocket-protocol", ", ".join(subprotocols))
        kwargs["headers"] = headers
        return headers

    def __enter__(self) -> TestClient:
        """
        Enters the context manager.

        Returns:
            TestClient: The test client.
        """
        with contextlib.ExitStack() as stack:
            self.portal = portal = stack.enter_context(
                anyio.from_thread.start_blocking_portal(**self.async_backend)
            )

            @stack.callback
            def reset_portal() -> None:
                self.portal = None

            send1: ObjectSendStream[MutableMapping[str, Any] | None]
            receive1: ObjectReceiveStream[MutableMapping[str, Any] | None]
            send2: ObjectSendStream[MutableMapping[str, Any]]
            receive2: ObjectReceiveStream[MutableMapping[str, Any]]
            send1, receive1 = anyio.create_memory_object_stream(math.inf)
            send2, receive2 = anyio.create_memory_object_stream(math.inf)
            self.stream_send = StapledObjectStream(send1, receive1)
            self.stream_receive = StapledObjectStream(send2, receive2)
            self.task = portal.start_task_soon(self.lifespan)
            portal.call(self.wait_startup)

            @stack.callback
            def wait_shutdown() -> None:
                portal.call(self.wait_shutdown)

            self.exit_stack = stack.pop_all()

        return self

    def __exit__(self, *args: Any) -> None:
        """
        Exits the context manager.
        """
        self.exit_stack.close()

    async def lifespan(self) -> None:
        """
        Handles the lifespan of the ASGI application.
        """
        scope = {"type": "lifespan", "state": self.app_state}
        try:
            await self.app(scope, self.stream_receive.receive, self.stream_send.send)
        finally:
            await self.stream_send.send(None)

    async def wait_startup(self) -> None:
        """
        Waits for the ASGI application to start up.
        """
        await self.stream_receive.send({"type": "lifespan.startup"})

        async def receive() -> Any:
            message = await self.stream_send.receive()
            if message is None:
                self.task.result()
            return message

        message = await receive()
        assert message["type"] in (
            "lifespan.startup.complete",
            "lifespan.startup.failed",
        )
        if message["type"] == "lifespan.startup.failed":
            await receive()

    async def wait_shutdown(self) -> None:
        """
        Waits for the ASGI application to shut down.
        """

        async def receive() -> Any:
            message = await self.stream_send.receive()
            if message is None:
                self.task.result()
            return message

        async with self.stream_send:
            await self.stream_receive.send({"type": "lifespan.shutdown"})
            message = await receive()
            assert message["type"] in (
                "lifespan.shutdown.complete",
                "lifespan.shutdown.failed",
            )
            if message["type"] == "lifespan.shutdown.failed":
                await receive()



from __future__ import annotations

import inspect
from collections.abc import Awaitable, Callable
from typing import TYPE_CHECKING, Any, cast

from lilya._internal._encoders import json_encode
from lilya._internal._exception_handlers import wrap_app_handling_exceptions
from lilya.compat import is_async_callable
from lilya.concurrency import run_in_threadpool
from lilya.context import Context
from lilya.enums import SignatureDefault
from lilya.requests import Request
from lilya.responses import Ok, Response
from lilya.types import ASGIApp, Receive, Scope, Send
from lilya.websockets import WebSocket

if TYPE_CHECKING:
    from lilya.routing import BasePath


class BaseHandler:
    """
    Utils to manage the responses of the handlers.
    """

    def handle_response(
        self, func: Callable[[Request], Awaitable[Response] | Response]
    ) -> ASGIApp:
        """
        Decorator for creating a request-response ASGI application.

        Args:
            func (Callable): The function to be wrapped.

        Returns:
            ASGIApp: The ASGI application.
        """

        async def app(scope: Scope, receive: Receive, send: Send) -> None:
            """
            ASGI application handling request-response.

            Args:
                scope (Scope): The request scope.
                receive (Receive): The receive channel.
                send (Send): The send channel.

            Returns:
                None
            """
            request = Request(scope=scope, receive=receive, send=send)

            async def inner_app(scope: Scope, receive: Receive, send: Send) -> None:
                """
                Inner ASGI application handling request-response.

                Sometimes the handler does not need the request to be passed
                in the handler and we can avoid it by ignoring the request
                object in the arguments.

                Args:
                    scope (Scope): The request scope.
                    receive (Receive): The receive channel.
                    send (Send): The send channel.

                Returns:
                    None
                """

                signature: inspect.Signature = self.signature
                func_params: dict[str, Any] = {
                    **self._extract_params_from_request(request=request, signature=signature),
                    **self._extract_context(request=request, signature=signature),
                }
                if signature.parameters:
                    if SignatureDefault.REQUEST in signature.parameters:
                        func_params.update({"request": request})
                        response = await self._execute_function(func, **func_params)
                    else:
                        response = await self._execute_function(func, **func_params)
                else:
                    response = await self._execute_function(func, **func_params)

                await self._handle_response_content(response, scope, receive, send)

            await wrap_app_handling_exceptions(inner_app, request)(scope, receive, send)

        return app

    async def _handle_response_content(
        self, app: ASGIApp, scope: Scope, receive: Receive, send: Send
    ) -> None:
        """
        Generates the app response, ensuring it is in the form of an ASGI application.
        When a special type is passed, it tries to convert to a json format and generate
        the response.

        Args:
            app (Union[ASGIApp, Any]): The response content.
            scope (Scope): The ASGI scope.
            receive (Receive): The receive channel.
            send (Send): The send channel.
        """
        if is_async_callable(app) or isinstance(app, Response):
            # If response is an ASGI application or an async callable, directly await it.
            await app(scope, receive, send)
        else:
            # If response is not an async callable, wrap it in an ASGI application and then await.
            if app is not None:
                app = json_encode(app)

            response = Ok(app)
            await response(scope, receive, send)

    def _extract_params_from_request(
        self, request: Request, signature: inspect.Signature
    ) -> dict[str, Any]:
        """
        Extracts parameters from the request and injects them into the function if needed.

        Args:
            request (Request): The incoming request.
            signature (inspect.Signature): The signature of the target function.

        Returns:
            Dict[str, Any]: A dictionary containing parameters extracted from the request.
        """
        return {
            param: value
            for param, value in request.path_params.items()
            if param in signature.parameters
        }

    def _extract_context(self, request: Request, signature: inspect.Signature) -> dict[str, Any]:
        """
        Extracts the context from the signature and injects them into the function if needed.

        Args:
            request (Request): The incoming request.
            signature (inspect.Signature): The signature of the target function.

        Returns:
            Dict[str, Any]: A dictionary containing parameters extracted from the signature.
        """
        params: dict[str, Any] = {}
        for param, _ in signature.parameters.items():
            if param in ("context",):
                value = Context(__handler__=cast("BasePath", self), __request__=request)
                params[param] = value
                break
        return params

    def handle_websocket_session(self, func: Callable[[WebSocket], Awaitable[None]]) -> ASGIApp:
        """
        Decorator for creating a WebSocket session ASGI application.

        Args:
            func (Callable): The function to be wrapped.

        Returns:
            ASGIApp: The ASGI application.
        """

        async def app(scope: Scope, receive: Receive, send: Send) -> None:
            """
            ASGI application handling WebSocket session.

            Args:
                scope (Scope): The request scope.
                receive (Receive): The receive channel.
                send (Send): The send channel.

            Returns:
                None
            """
            session = WebSocket(scope=scope, receive=receive, send=send)

            async def inner_app(scope: Scope, receive: Receive, send: Send) -> None:
                """
                Inner ASGI application handling WebSocket session.

                Args:
                    scope (Scope): The request scope.
                    receive (Receive): The receive channel.
                    send (Send): The send channel.

                Returns:
                    None
                """
                await self._execute_function(func, session)

            await wrap_app_handling_exceptions(inner_app, session)(scope, receive, send)

        return app

    async def _execute_function(self, func: Callable, *args: Any, **kwargs: Any) -> Any:
        """
        Executes the given function, handling both synchronous and asynchronous functions.

        Args:
            func (Callable): The function to execute.
            args (Any): Positional arguments for the function.
            kwargs (Any): Keyword arguments for the function.

        Returns:
            Any: The result of the function execution.
        """
        if is_async_callable(func):
            return await func(*args, **kwargs)
        else:
            return await run_in_threadpool(func, *args, **kwargs)



from __future__ import annotations

import re
from collections import namedtuple
from collections.abc import Iterable
from re import Pattern
from typing import Any, TypeVar, cast

from lilya._internal._path_transformers import (
    TRANSFORMER_PYTHON_TYPES,
    TRANSFORMER_TYPES,
    Transformer,
)
from lilya.types import Scope

T = TypeVar("T")

PathParameter = namedtuple("PathParameter", ["name", "type"])


def clean_path(path: str) -> str:
    """
    Make sure a given path by ensuring it starts with a slash and does not
    end with a slash.
    """
    path = path.rstrip("/")
    if not path.startswith("/"):
        path = "/" + path
    path = re.sub("//+", "/", path)
    return path


def remove_start_slashes(path: str) -> str:
    """
    Removes all the beginning starting with slashes
    """
    path = path.lstrip("/")
    return path


def join_paths(paths: Iterable[str]) -> str:
    return clean_path("/".join(paths))


def get_route_path(scope: Scope) -> str:
    root_path = scope.get("root_path", "")
    return cast(
        str,
        (
            scope["path"][len(root_path) :]
            if root_path and scope["path"].startswith(root_path)
            else scope["path"]
        ),
    )


def replace_params(
    path: str,
    param_convertors: dict[str, Transformer[Any]],
    path_params: dict[str, str],
    is_host: bool = False,
) -> tuple[str, dict[str, str]]:
    """
    Replaces placeholders in the given path with normalized values from path parameters.

    Args:
        path (str): The path string with placeholders.
        param_convertors (Dict[str, Transformer[Any]]): Dictionary of parameter names and transformers.
        path_params (Dict[str, str]): Dictionary of path parameters.

    Returns:
        Tuple[str, Dict[str, str]]: The updated path and any remaining path parameters.
    """
    updated_path = path
    remaining_params = {}
    for key, value in list(path_params.items()):
        placeholder = "{" + key + "}"

        if placeholder in updated_path:
            transformer = param_convertors[key]
            normalized_value = transformer.normalise(value)
            updated_path = updated_path.replace(placeholder, normalized_value)
            path_params.pop(key)
        else:
            remaining_params[key] = value

    if not is_host:
        return clean_path(updated_path), remaining_params
    return updated_path, remaining_params


def compile_path(path: str) -> tuple[Pattern[str], str, dict[str, Transformer[Any]], str]:
    """
    Compile a path or host string into a three-tuple of (regex, format, {param_name:convertor}).

    Args:
        path (str): The path or host string.

    Returns:
        Tuple[Pattern[str], str, Dict[str, 'Transformer[Any]']]: The compiled regex pattern, format string,
        and a dictionary of parameter names and converters.
    """
    is_host = not path.startswith("/")
    path_regex, path_format, param_convertors, path_start = generate_regex_and_format(
        path, is_host
    )

    return re.compile(path_regex), path_format, param_convertors, path_start


def generate_regex_and_format(
    path: str, is_host: bool
) -> tuple[str, str, dict[str, Transformer[Any]], str]:
    path_regex, path_format, param_convertors, index = "^", "", {}, 0  # type: ignore
    duplicate_params: set[str] = set()

    for match in re.finditer(r"[\{\<]([a-zA-Z_]\w*)(:[a-zA-Z_]\w*)?[\}\>]", path):
        param_name, convertor_type = match.groups("str")
        convertor_type = convertor_type.lstrip(":")

        convertor = get_transformer(convertor_type)
        (
            path_regex,
            path_format,
            param_convertors,
            index,
            path_start,
            duplicate_param,
        ) = update_paths_and_convertors(
            path,
            path_regex,
            path_format,
            param_convertors,
            index,
            param_name,
            convertor,
            match,
        )
        if duplicate_param:
            duplicate_params.add(duplicate_param)
    else:
        path_start = path

    if is_host:
        hostname = extract_hostname(path, index)
        path_regex += re.escape(hostname) + "$"
    else:
        path_regex += re.escape(path[index:]) + "$"

    raise_for_duplicate_params(path, duplicate_params)
    path_format += path[index:]
    return path_regex, path_format, param_convertors, path_start


def get_transformer(transformer_type: str) -> Transformer[Any]:
    assert transformer_type in TRANSFORMER_TYPES, f"Unknown path transformer '{transformer_type}'"
    return TRANSFORMER_TYPES[transformer_type]


def update_paths_and_convertors(
    path: str,
    path_regex: str,
    path_format: str,
    param_convertors: dict[str, Transformer[Any]],
    index: int,
    param_name: str,
    convertor: Transformer[Any],
    match: re.Match,
) -> tuple[str, str, dict[str, Transformer[Any]], int, str, str]:
    path_start = path[index : match.start()]

    path_regex += re.escape(path_start)
    path_regex += f"(?P<{param_name}>{convertor.regex})"

    path_format += path_start
    path_format += f"{{{param_name}}}"
    duplicate_param: str = None

    if param_name in param_convertors:
        duplicate_param = param_name

    param_convertors[param_name] = convertor
    index = match.end()

    return path_regex, path_format, param_convertors, index, path_start, duplicate_param


def raise_for_duplicate_params(path: str, duplicate_params: set[str] | None = None) -> None:
    """
    Builds and generates the error message for duplicate parameters
    in the path.
    """
    if not duplicate_params:
        return

    names = ", ".join(sorted(duplicate_params))
    ending = "s" if len(duplicate_params) > 1 else ""
    raise ValueError(f"Duplicated param name{ending} {names} in the path {path}")


def extract_hostname(path: str, index: int) -> str:
    return path[index:].split(":")[0]


def parse_path(path: str) -> list[str | Any]:
    """
    Using the TRANSFORMERS and the application registered convertors,
    transforms the path into definition.
    """
    _, path, variables, _ = compile_path(path)

    parsed_components: list[str | Any] = []

    for name, transformer in variables.items():
        _type = TRANSFORMER_PYTHON_TYPES[transformer.__class__.__name__]
        parsed_components.append(PathParameter(name=name, type=_type))
    return parsed_components



from __future__ import annotations

from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from lilya.testclient._internal.websockets import WebSocketTestSession


class UpgradeException(Exception):
    def __init__(self, session: WebSocketTestSession) -> None:
        self.session = session


class ASGISpecViolation(Exception): ...



from __future__ import annotations

import json
from base64 import b64decode, b64encode
from collections.abc import Callable
from typing import Any, Literal

import itsdangerous
from itsdangerous.exc import BadSignature

from lilya._internal._connection import Connection
from lilya.datastructures import Header, Secret
from lilya.protocols.middleware import MiddlewareProtocol
from lilya.types import ASGIApp, Message, Receive, Scope, Send


class SessionMiddleware(MiddlewareProtocol):
    def __init__(
        self,
        app: ASGIApp,
        secret_key: str | Secret,
        session_cookie: str = "session",
        max_age: int | None = 14 * 24 * 60 * 60,  # 14 days, in seconds
        path: str = "/",
        same_site: Literal["lax", "strict", "none"] = "lax",
        https_only: bool = False,
        domain: str | None = None,
        session_serializer: Callable[[Any], bytes | str] = json.dumps,
        session_deserializer: Callable[[bytes], Any] = json.loads,
    ) -> None:
        """
        Middleware for handling session data in ASGI applications.

        Args:
            app (ASGIApp): The ASGI application to wrap.
            secret_key (Union[str, Secret]): The secret key used for signing session data.
            session_cookie (str): The name of the session cookie.
            max_age (Optional[int]): The maximum age of the session in seconds (default is 14 days).
            path (str): The path attribute for the session cookie.
            same_site (Literal["lax", "strict", "none"]): The SameSite attribute for the session cookie.
            https_only (bool): If True, set the secure flag for the session cookie (HTTPS only).
            domain (Optional[str]): The domain attribute for the session cookie.
            session_serializer (Callable[[Any], bytes | str]): The encoder for the session. Default json.dumps.
            session_deserializer (Callable[[bytes], Any]): The decoder for the session. Default json.loads.
        """
        self.app = app
        self.signer = itsdangerous.TimestampSigner(str(secret_key))
        self.session_cookie = session_cookie
        self.max_age = max_age
        self.path = path
        self.security_flags = "httponly; samesite=" + same_site
        if https_only:
            self.security_flags += "; secure"
        if domain is not None:
            self.security_flags += f"; domain={domain}"
        self.session_serializer = session_serializer
        self.session_deserializer = session_deserializer

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        ASGI application callable.

        Args:
            scope (Scope): ASGI scope.
            receive (Receive): ASGI receive channel.
            send (Send): ASGI send channel.
        """
        if scope["type"] not in ("http", "websocket"):
            await self.app(scope, receive, send)
            return

        connection = Connection(scope)
        initial_session_was_empty = await self.load_session_data(scope, connection)

        async def send_wrapper(message: Message) -> None:
            await self.process_response(message, scope, initial_session_was_empty, send)

        await self.app(scope, receive, send_wrapper)

    def decode_session(self, data: bytes) -> Any:
        return self.session_deserializer(b64decode(data))

    def encode_session(self, session: Any) -> bytes:
        data = self.session_serializer(session)
        if isinstance(data, str):
            data = data.encode("utf-8")
        return b64encode(data)

    async def load_session_data(self, scope: Scope, connection: Connection) -> bool:
        """
        Load session data from the session cookie.

        Args:
            scope (Scope): ASGI scope.
            connection (Connection): HTTP connection object.

        Returns:
            bool: True if the initial session was empty, False otherwise.
        """
        if self.session_cookie in connection.cookies:
            data = connection.cookies[self.session_cookie].encode("utf-8")
            try:
                data = self.signer.unsign(data, max_age=self.max_age)
                scope["session"] = self.decode_session(data)
                return False
            except BadSignature:
                scope["session"] = {}
        else:
            scope["session"] = {}
        return True

    async def process_response(
        self,
        message: Message,
        scope: Scope,
        initial_session_was_empty: bool,
        send: Send,
    ) -> None:
        """
        Process the response and set the session cookie.

        Args:
            message (Message): ASGI message.
            scope (Scope): ASGI scope.
            initial_session_was_empty (bool): True if the initial session was empty, False otherwise.
            send (Send): ASGI send channel.
        """
        if message["type"] == "http.response.start":
            if scope["session"]:
                message = await self.set_session_cookie(scope, message)
            elif not initial_session_was_empty:
                message = await self.clear_session_cookie(scope, message)

        await send(message)

    async def set_session_cookie(self, scope: Scope, message: Message) -> Message:
        """
        Set the session cookie in the response headers.

        Args:
            scope (Scope): ASGI scope.
            message (Message): ASGI message
        """
        data = self.encode_session(scope["session"])
        data = self.signer.sign(data)
        headers = Header.from_scope(scope=scope)
        header_value = "{session_cookie}={data}; path={path}; {max_age}{security_flags}".format(
            session_cookie=self.session_cookie,
            data=data.decode("utf-8"),
            path=self.path,
            max_age=f"Max-Age={self.max_age}; " if self.max_age else "",
            security_flags=self.security_flags,
        )
        headers.add("Set-Cookie", header_value)
        message["headers"] = headers.encoded_multi_items()
        return message

    async def clear_session_cookie(self, scope: Scope, message: Message) -> Message:
        """
        Clear the session cookie in the response headers.

        Args:
            scope (Scope): ASGI scope.
        """
        headers = Header.from_scope(scope=scope)
        header_value = "{session_cookie}={data}; path={path}; {expires}{security_flags}".format(
            session_cookie=self.session_cookie,
            data="null",
            path=self.path,
            expires="expires=Thu, 01 Jan 1970 00:00:00 GMT; ",
            security_flags=self.security_flags,
        )
        headers.add("Set-Cookie", header_value)
        message["headers"] = headers.encoded_multi_items()
        return message



from __future__ import annotations

import contextvars
import copy
import warnings
from functools import lru_cache
from typing import TYPE_CHECKING, Annotated, Any, cast

from lilya.datastructures import URL
from lilya.exceptions import ImproperlyConfigured
from lilya.types import Doc, Scope

if TYPE_CHECKING:
    from lilya.requests import Request
    from lilya.routing import BasePath


class Context:
    """
    `Context` class is used for the handlers context of the
    scope of the call.

    When a `context` is passed through the handler,
    the context will be aware of the decoractor handler
    itself.

    You will probably not need the context or change it but it is here in
    case you decide to use it.

    !!! Tip
        The context only exists in the handlers and nothing else which you can also
        see it as `request context` sort of approach.

    **Example**

    ```python
    from lilya.apps import Lilya
    from lilya.requests import Request
    from lilya.context import Context


    async def home(context: Context) -> None: ...
    ```
    """

    def __init__(
        self,
        __handler__: Annotated[
            BasePath,
            Doc(
                """
                The [handler](https://lilya.dev/routing) where the context will be. placed.

                To avoid any adulteration of the the original route handler, the context performs
                a shallow copy of the original handler itself.
                """
            ),
        ],
        __request__: Annotated[
            Request,
            Doc(
                """
                A [Request](https://lilya.dev/references/request/) class object.
                """
            ),
        ],
    ) -> None:
        self.__handler__ = copy.copy(__handler__)
        self.__request__ = __request__

    @property
    def handler(self) -> BasePath:
        return self.__handler__

    @property
    def request(self) -> Request:
        return self.__request__

    @property
    def user(self) -> Any:
        return self.request.user

    @property
    def settings(self) -> Any:
        return self.request.app.settings

    @property
    def scope(self) -> Scope:
        return self.request.scope

    @property
    def app(self) -> Any:
        return self.request.scope["app"]

    def add_to_context(
        self,
        key: Annotated[
            Any,
            Doc(
                """
                The key value to be added to the context dictionary.
                """
            ),
        ],
        value: Annotated[
            Any,
            Doc(
                """
                The value value to be added to the context dictionary and map with the key.
                """
            ),
        ],
    ) -> None:
        """
        Adds a key to the context of an handler.

        This can be particularly useful if you are programatically
        building the handler or for any other specific and unique operation.

        **Example**

        ```python
        from typing import Any
        from lilya.context import Context

        async def get_data(context: Context) -> Any:
            context.update{"name": "Lilya"}
            return context.get_context_data()
        ```
        """
        if key in self.__dict__:
            warnings.warn(
                f"The key: '{key} already exists in context and it will be overritten.",
                stacklevel=2,
            )
        setattr(self, key, value)

    def get_context_data(self, **kwargs: Any) -> dict[Any, Any]:
        """
        Returns the context in a python dictionary like structure.
        """
        context_data = {
            k: v
            for k, v in self.__dict__.items()
            if not k.startswith("__") and not k.endswith("__")
        }
        return context_data

    def path_for(
        self,
        name: Annotated[
            str,
            Doc(
                """
                The `name` given in a `Gateway` or `Include` objects.

                **Example**

                ```python
                from lilya import Lilya, Gateway

                Gateway(handler=..., name="view-users")
                ```
                """
            ),
        ],
        **path_params: Annotated[
            Any,
            Doc(
                """
                Any additional *path_params* declared in the URL of the
                handler and are needed to *reverse* the names and return
                the proper URL.
                """
            ),
        ],
    ) -> str:
        url: URL = self.request.path_for(name, **path_params)
        return str(url)


class G:
    """
    This object represents a global state of the application.
    This can be used for Lilya requests and to store global
    properties on a request context of the application and share
    the state across.

    This diverges from `state` of the application object and state
    that is only accessible within the handler itself.
    """

    _context: dict[str, Any]

    def __init__(self, context: dict[str, Any] | None = None):
        if context is None:
            context = {}
        super().__setattr__("_context", context)

    @property
    def store(self) -> Any:
        return self._context

    def __setattr__(self, key: Any, value: Any) -> None:
        self._context[key] = value

    def __getattr__(self, key: Any) -> Any:
        try:
            return self._context[key]
        except KeyError:
            message = "'{}' object has no attribute '{}'"
            raise AttributeError(message.format(self.__class__.__name__, key)) from None

    def __delattr__(self, key: Any) -> None:
        del self._context[key]

    def __copy__(self) -> G:
        return self.__class__(copy.copy(self._context))

    def __len__(self) -> int:
        return len(self._context)

    def __getitem__(self, key: str) -> Any:
        return self._context[key]

    def copy(self) -> G:
        return copy.copy(self)

    def clear(self) -> None:
        self._context.clear()

    def __repr__(self: G) -> str:
        return f"{self.__class__.__name__}()"


@lru_cache
def get_g() -> G:
    return G()


g = get_g()


class LazyRequestProxy:
    """
    A proxy to lazily evaluate attributes of the current request.
    """

    def __init__(self, request_getter: Any) -> None:
        self._request_getter = request_getter

    def __getattr__(self, item: str) -> Any:
        request = self._request_getter()
        if not request:
            raise ImproperlyConfigured(
                "Request context requires the 'RequestContextMiddleware' to be installed."
            )
        return getattr(request, item)

    def __repr__(self: LazyRequestProxy) -> str:
        if self._request_getter is None:
            return "<LazyRequestProxy [Unevaluated]>"
        return '<LazyRequestProxy "Request()">'


class RequestContext:
    """
    Similar to the request, this object makes it available to the handler
    to be access anywhere in the application that does not require the handler directly.
    """

    _request_context: contextvars.ContextVar[Request] = contextvars.ContextVar("request_context")

    # Lazily access the request object
    lazy_request = LazyRequestProxy(lambda: RequestContext._request_context.get(None))

    @classmethod
    def set_request(cls, request: Request) -> None:
        """
        Set the current request in the context.
        """
        cls._request_context.set(request)

    @classmethod
    def get_request(cls) -> Request:
        """
        Retrieve the current request from the context.
        """
        try:
            return cls._request_context.get()
        except LookupError:
            raise RuntimeError(
                "Request context requires the 'RequestContextMiddleware' to be installed."
            ) from None

    @classmethod
    def reset_request(cls, token: Any) -> None:
        """
        Reset the context variable.
        """
        cls._request_context.reset(token)

    def __str__(self) -> str:
        return "<Request()>"

    def __repr__(self) -> str:
        return str(self)


request_context: Request = cast("Request", RequestContext.lazy_request)


class LazySessionProxy:
    def __init__(self, session_getter: Any) -> None:
        self._session_getter = session_getter

    def __getitem__(self, key: str) -> Any:
        session = self._session_getter()
        if session is None:
            raise RuntimeError("Session context requires 'SessionMiddleware' to be installed.")
        return session[key]

    def __setitem__(self, key: str, value: Any) -> None:
        session = self._session_getter()
        if session is None:
            raise RuntimeError("Session context requires 'SessionMiddleware' to be installed.")
        session[key] = value

    def __delitem__(self, key: str) -> None:
        session = self._session_getter()
        if session is None:
            raise RuntimeError("Session context requires 'SessionMiddleware' to be installed.")
        del session[key]

    def __contains__(self, key: str) -> bool:
        session = self._session_getter()
        return key in session if session else False

    def get(self, key: str, default: Any = None) -> Any:
        session = self._session_getter()
        return session.get(key, default) if session else default

    def reset_context(self) -> None:
        session = self._session_getter()
        session.clear()

    def __repr__(self: LazySessionProxy) -> str:
        return '<LazySessionProxy "session">'


class SessionContext:
    """
    A context manager for handling session-specific request data using context variables.
    """

    _session_context: contextvars.ContextVar[Request] = contextvars.ContextVar("session_context")

    # Lazily access the session object
    lazy_session = LazySessionProxy(lambda: SessionContext._get_session_data())

    @classmethod
    def set_request(cls, request: Request) -> None:
        """Set the session context for the request."""
        cls._session_context.set(request)

    @classmethod
    def get_request(cls) -> Request:
        """Retrieve the current request (session context)."""
        try:
            return cls._session_context.get()
        except LookupError:
            raise RuntimeError(
                "Session context not set. Ensure 'SessionMiddleware' is properly installed."
            ) from None

    @classmethod
    def _get_session_data(cls) -> Any:
        """Get the session data from the request."""
        request = cls.get_request()
        return getattr(request, "session", {})

    @classmethod
    def reset_context(cls, token: Any) -> None:
        """
        Reset the session context.
        """
        # Ensure the token is a valid instance of Token
        if not isinstance(token, contextvars.Token):
            raise TypeError(f"Expected a Token, got {type(token)}")
        cls._session_context.reset(token)


# Define session as a lazy proxy
session: LazySessionProxy = SessionContext.lazy_session



from __future__ import annotations

from abc import ABC

from lilya.context import g
from lilya.protocols.middleware import MiddlewareProtocol
from lilya.types import ASGIApp, Receive, Scope, Send


class GlobalContextMiddleware(ABC, MiddlewareProtocol):
    """
    GlobalContextMiddleware is an ASGI middleware that initializes a global context for each request.

    This middleware inherits from ABC and MiddlewareProtocol, ensuring it adheres to the ASGI middleware interface.

    Attributes:
        app (ASGIApp): The ASGI application instance.

    Methods:
        __init__(app: ASGIApp):
            Initializes the middleware with the given ASGI application.

        __call__(scope: Scope, receive: Receive, send: Send) -> None:
            Asynchronously handles the incoming request, sets up a new global context,
            and then calls the next middleware or application in the stack.

    Usage:
        This middleware should be added to the ASGI application stack to ensure that
        a fresh global context is available for each request. This can be useful for
        storing request-specific data that needs to be accessed globally within the
        request lifecycle.
    """

    def __init__(self, app: ASGIApp) -> None:
        super().__init__(app)
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Asynchronous call method to set up the global context and then call the parent class's __call__ method.

        Args:
            scope (Scope): The scope of the ASGI application.
            receive (Receive): The receive function to get messages from the client.
            send (Send): The send function to send messages to the client.

        Returns:
            None
        """
        try:
            await self.app(scope, receive, send)
        finally:
            g.clear()



from __future__ import annotations

import os
import sys
from typing import Any, cast

import click

from lilya.cli.env import DirectiveEnv
from lilya.cli.exceptions import DirectiveError
from lilya.cli.terminal import OutputColour, Print, Terminal

printer = Print()
terminal = Terminal()


@click.option(
    "-p",
    "--port",
    type=int,
    default=8000,
    help="Port to run the development server.",
    show_default=True,
)
@click.option(
    "-r",
    "--reload",
    type=bool,
    default=True,
    help="Reload server on file changes.",
    is_flag=True,
    show_default=True,
)
@click.option(
    "--host",
    type=str,
    default="localhost",
    help="Server host. Tipically localhost.",
    show_default=True,
)
@click.option(
    "--debug",
    type=bool,
    default=True,
    help="Start the application in debug mode.",
    show_default=True,
    is_flag=True,
)
@click.option(
    "--log-level",
    type=str,
    default="debug",
    help="What log level should uvicorn run.",
    show_default=True,
)
@click.option(
    "--lifespan",
    type=str,
    default="on",
    help="Enable lifespan events.",
    show_default=True,
)
@click.option(
    "--settings", type=str, help="Any custom settings to be initialised.", required=False
)
@click.command(name="runserver")
def runserver(
    env: DirectiveEnv,
    port: int,
    reload: bool,
    host: str,
    debug: bool,
    log_level: str,
    lifespan: str,
    settings: str | None = None,
) -> None:
    """Starts the Lilya development server.

    The --app can be passed in the form of <module>.<submodule>:<app> or be set
    as environment variable LILYA_DEFAULT_APP.

    Alternatively, if none is passed, Lilya will perform the application discovery.

    It is strongly advised not to run this command in any pther environment but developmentyping.
    This was designed to facilitate the development environment and should not be used in pr

    How to run: `lilya runserver`
    """
    if getattr(env, "app", None) is None:
        error = (
            "You cannot specify a custom directive without specifying the --app or setting "
            "LILYA_DEFAULT_APP environment variable."
        )
        printer.write_error(error)
        sys.exit(1)

    if settings is not None:
        os.environ.setdefault("LILYA_SETTINGS_MODULE", settings)

    try:
        import uvicorn
    except ImportError:
        raise DirectiveError(detail="Uvicorn needs to be installed to run Lilya.") from None

    server_environment: str = ""
    if os.environ.get("LILYA_SETTINGS_MODULE"):
        from lilya.conf import settings as lilya_settings

        server_environment = f"{lilya_settings.environment} "

    app = env.app
    message = terminal.write_info(
        f"Starting {server_environment}server @ {host} server @ {host}",
        colour=OutputColour.BRIGHT_CYAN,
    )
    terminal.rule(message, align="center")

    if os.environ.get("LILYA_SETTINGS_MODULE"):
        custom_message = f"'{os.environ.get('LILYA_SETTINGS_MODULE')}'"
        terminal.rule(custom_message, align="center")

    if debug:
        app.debug = debug

    uvicorn.run(
        app=env.path,
        port=port,
        host=host,
        reload=reload,
        lifespan=cast(Any, lifespan),
        log_level=log_level,
    )



from __future__ import annotations

from lilya.context import SessionContext
from lilya.protocols.middleware import MiddlewareProtocol
from lilya.requests import Request
from lilya.types import ASGIApp, Receive, Scope, Send


class SessionContextMiddleware(MiddlewareProtocol):
    """
    Middleware to manage session context in an ASGI application.
    """

    def __init__(self, app: ASGIApp) -> None:
        """
        Initializes the SessionContextMiddleware with the given ASGI application.
        """
        self.app = app

    async def __call__(self, scope: Scope, receive: Receive, send: Send) -> None:
        """
        Handles incoming requests, sets the session context, and resets after request processing.
        """
        if scope["type"] != "http":
            await self.app(scope, receive, send)
            return

        request = Request(scope)

        # Initialize session if not already initialized
        if not hasattr(request, "session"):
            request.session = {}

        # Set the session context
        token = SessionContext._session_context.set(request)
        try:
            await self.app(scope, receive, send)
        finally:
            SessionContext.reset_context(token)



from __future__ import annotations

import sys
from functools import wraps
from typing import Any

from lilya.conf import settings
from lilya.conf.context_vars import set_override_settings
from lilya.conf.global_settings import Settings

if sys.version_info >= (3, 10):  # pragma: no cover
    from typing import ParamSpec
else:  # pragma: no cover
    from typing_extensions import ParamSpec

P = ParamSpec("P")


class override_settings:
    """
    A context manager that allows overriding Lilya settings temporarily.

    Usage:
    ```
    with override_settings(SETTING_NAME=value):
        # code that uses the overridden settings
    ```

    The `override_settings` class can also be used as a decorator.

    Usage:
    ```
    @override_settings(SETTING_NAME=value)
    def test_function():
        # code that uses the overridden settings
    ```
    """

    def __init__(self, **kwargs: Any) -> None:
        """
        Initialize the class with the given keyword arguments.

        Args:
            **kwargs: Additional keyword arguments to be stored as options.

        Returns:
            None
        """
        self.app = kwargs.pop("app", None)
        self.options = kwargs

    def __enter__(self) -> None:
        """
        Enter the context manager and set the modified settings.

        Saves the original settings and sets the modified settings
        based on the provided options.

        Returns:
            None
        """
        self._original_settings = settings._wrapped
        settings._wrapped = Settings(settings._wrapped, **self.options)
        set_override_settings(True)

    def __exit__(self, exc_type: Any, exc_value: Any, traceback: Any) -> None:
        """
        Restores the original settings and sets them up again.

        Args:
            exc_type (Any): The type of the exception raised, if any.
            exc_value (Any): The exception instance raised, if any.
            traceback (Any): The traceback for the exception raised, if any.
        """
        settings._wrapped = self._original_settings
        settings._setup()
        set_override_settings(False)

    def __call__(self, test_func: Any) -> Any:
        """
        Decorator that wraps a test function and executes it within a context manager.

        Args:
            test_func (Any): The test function to be wrapped.

        Returns:
            Any: The result of the test function.

        """

        @wraps(test_func)
        def inner(*args: P.args, **kwargs: P.kwargs) -> Any:
            with self:
                return test_func(*args, **kwargs)

        return inner
